package scimark2;

/**
	SciMark2: A Java numerical benchmark measuring performance
	of computational kernels for FFTs, Monte Carlo simulation,
	sparse matrix computations, Jacobi SOR, and dense LU matrix
	factorizations.  

	<p>
	Units returned are  number of iterations per second
	for each kernel.  (Roughly calibrated on a Sun SPARC 10,
	using the JDK 1.1.7.) The SciMark composite score is
	the average of measurements for each kernel.

	<p>
	(See http://math.nist.gov/scimark for details.)

	<p>
	To run program use:

	<pre>
	>java scimark2
	</pre>

*/

import java.util.Random;

public class scimark2
{

  /* perform 6 variants of matrix multiply and return a 
  	 composite Mflop results for all results.

	 results[0] has the composite score.

  */


	public static final double RESOLUTION_DEFAULT = 2.0;
	private static final int FFT_SIZE = 4096;  // must be a power of two
	private static final int JACOBI_SIZE =100; // NxN grid
	private static final int JACOBI_NUM_ITER = 20; // number of iterations
											   	//   (for granularity)
	private static final int MONTECARLO_SIZE = 25000; 
	private static final int SPARSE_SIZE_M = 5000;
	private static final int SPARSE_SIZE_N = 5000;
	private static final int SPARSE_SIZE_nz = 25000;
	private static final int SPARSE_NUM_ITER = 10;
	private static final int LU_SIZE = 100;

	// seed needs to be explicitly set for reproducibility
	// (not critical what actual value is used, as long as it is consistent.
	//
	private static final long RANDOM_SEED = 10101010;  

	public static void main(String args[])
	{
		double resolution_time = RESOLUTION_DEFAULT;

        if (args.length > 0)
        {
        	resolution_time = Double.valueOf(args[0]).doubleValue();
        }
        

		double res[] = benchmark(resolution_time);

		System.out.println("Composite Score: " + res[0]);
		System.out.println("FFT ("+FFT_SIZE+"): " + res[1]);
		System.out.println("Jacobi SOR ("+JACOBI_SIZE+"x"+JACOBI_SIZE+"): "
				+ "  " + res[2]);
		System.out.println("Monte Carlo ("+MONTECARLO_SIZE+"): " + res[3]);
		System.out.println("Sparse matmult (nz=" + SPARSE_SIZE_nz +
				"), 10 iterations: " + res[4]);
		System.out.println("LU (" + LU_SIZE + "x" + LU_SIZE + "): " + res[5]);

	}


  public double[] execute() 

  {
  	/* use a time granularity of 1.0 seconds */
	return benchmark(RESOLUTION_DEFAULT);
  }


	// return randon (M,N) matrix with entries between 0.0 and 1.0

  private static double[][] RandomMatrix(int M, int N, java.util.Random R)
  {
  		double A[][] = new double[M][N];

        for (int i=0; i<N; i++)
			for (int j=0; j<N; j++)
			{
            	A[i][j] = R.nextDouble() * 1e-6;
			}	
		return A;
	}

	private static double[] RandomVector(int N, java.util.Random R)
	{
		double A[] = new double[N];

		for (int i=0; i<N; i++)
			A[i] = R.nextDouble() * 1e-6;

		return A;
	}

  public static double[] benchmark(double TIME_LENGTH)
  {
  		// returns elapsed time for one iteration 

  		final int NUM_BENCHMARKS = 5;	
		// first result is the composite score
		//
  		double results[] = new double[NUM_BENCHMARKS+1];
		
		Random R = new Random(RANDOM_SEED);

		
		// preload classes before timing...

			// (1) touch the FFT routine
			double y1[] = RandomVector(2*16, R);
			FFT.transform(y1);


		Stopwatch Q = new Stopwatch();
                
		long cycles=1;
		int exp_num = 1;		


        /* --------------*/
		/* FFT benchmark */
        /* --------------*/



		// FFT_SIZE is the number of complex numbers used
		// (we use pairs of real/imag doubles)

	   {
		double x[] = RandomVector(2*(FFT_SIZE),R);


		cycles = 1;
		while(true)
		{
			Q.start();
			for (int i=0; i<cycles; i++)
			{
				FFT.transform(x);	// forward transform
				FFT.inverse(x);		// backward transform
			}
			Q.stop();
			if (Q.read() >= TIME_LENGTH) break;

			cycles *= 2;
		}
		// approx Mflops
		results[exp_num] = FFT.num_flops(FFT_SIZE)*cycles/ Q.read() * 1.0e-6;
		exp_num++;
		}
		System.gc();


		/******************/
		/*    Jacobi SOR  */
		/******************/

		
		{		
		double G[][] = RandomMatrix(JACOBI_SIZE, JACOBI_SIZE,R);

		cycles=1;
		while(true)
		{
			Q.start();
			for (int i=0; i<cycles; i++)
				Jacobi.SOR(1.25, G, JACOBI_NUM_ITER);
			Q.stop();
			if (Q.read() >= TIME_LENGTH) break;

			cycles *= 2;
		}
		// approx Mflops
		results[exp_num] = Jacobi.num_flops(JACOBI_SIZE, JACOBI_SIZE,
					JACOBI_NUM_ITER)* cycles / Q.read() * 1.0e-6;
		exp_num++;
		}
		System.gc();

		/****************************/
		/*  Monte Carlo integration */
		/****************************/


		cycles=1;
		while(true)
		{
			Q.start();
			for (int i=0; i<cycles; i++)
				MonteCarlo.integrate(MONTECARLO_SIZE);
			Q.stop();
			if (Q.read() >= TIME_LENGTH) break;

			cycles *= 2;
		}
		// approx Mflops
		//
		results[exp_num] = MonteCarlo.num_flops(MONTECARLO_SIZE)
				* cycles / Q.read() * 1.0e-6;
		exp_num++;
		System.gc();


		/***********************/
		/*    Sparse Matmult   */
		/***********************/

		{
		double x[] = RandomVector(SPARSE_SIZE_N, R);
		double y[] = new double[SPARSE_SIZE_M];

		double val[] = new double[SPARSE_SIZE_nz];
		int col[] = new int[SPARSE_SIZE_nz];
		int row[] = new int[SPARSE_SIZE_nz];

		/* initalize sparse matrix */

            for (int i=0; i<SPARSE_SIZE_nz; i++)
			{
				// generate random row index (0, M-1)
				row[i] = Math.abs(R.nextInt()) % SPARSE_SIZE_M;

				// generate random column index (0, N-1)
				col[i] = Math.abs(R.nextInt()) % SPARSE_SIZE_N;

				val[i] = R.nextDouble();
			}



		cycles=1;
		while(true)
		{
			Q.start();
			for (int c=0; c<cycles; c++)
			{
				/* mat-vec kernel */

				SparseMatmult.test(y, val, row, col, x, 
					SPARSE_NUM_ITER);
    
			}
			
			Q.stop();
			if (Q.read() >= TIME_LENGTH) break;

			cycles *= 2;
		}
		// approx Mflops
		results[exp_num] = SparseMatmult.num_flops(
					SPARSE_SIZE_nz, SPARSE_NUM_ITER) * cycles/ Q.read() 
					* 1.0e-6;
		exp_num++;
		}
		System.gc();

		//System.gc();


		/********************/
		/*   Dense LU       */
		/********************/

		
		{		
		double A[][] = RandomMatrix(LU_SIZE, LU_SIZE, R);

		cycles=1;
		while(true)
		{
			Q.start();
			for (int i=0; i<cycles; i++)
			{
				LU tmp = new LU(A);
			}
			Q.stop();
			if (Q.read() >= TIME_LENGTH) break;

			cycles *= 2;
		} 
		// approx Mflops
		//
		results[exp_num] = LU.num_flops(LU_SIZE)*cycles/ Q.read() * 1.0e-6;
		exp_num++;
		}
		System.gc();


		// compute composite score
		//
		double sum =0.0;
		for (int i=1; i<=NUM_BENCHMARKS; i++)
		{
			sum  += results[i];
		}

		results[0] = sum / NUM_BENCHMARKS;
		return results;
   }
  
}
